// Copyright (c) 2022-2024 INRIA Sophia-Antipolis (France), GeometryFactory (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Isosurfacing_3/include/CGAL/Isosurfacing_3/Octree_partition.h $
// $Id: include/CGAL/Isosurfacing_3/Octree_partition.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Sven Oesau

#ifndef CGAL_ISOSURFACING_3_OCTREE_PARTITION_H
#define CGAL_ISOSURFACING_3_OCTREE_PARTITION_H

#include <CGAL/license/Isosurfacing_3.h>

#include <CGAL/Octree.h>
#include <CGAL/Isosurfacing_3/internal/Octree_domain_3.h>
#include <CGAL/Isosurfacing_3/internal/partition_traits_Octree.h>

#endif // CGAL_ISOSURFACING_3_OCTREE_PARTITION_H
