// Copyright (c) 2010 GeometryFactory
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Number_types/include/CGAL/Quotient_fwd.h $
// $Id: include/CGAL/Quotient_fwd.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Sebastien Loriot, Laurent Rineau


#ifndef CGAL_QUOTIENT_FWD_H
#define CGAL_QUOTIENT_FWD_H

namespace CGAL{
  template < typename NT >
  struct Split_double;

  template <class NT_>
  class Quotient;

} //namespace CGAL

#endif //CGAL_QUOTIENT_FWD_H
