// Copyright (c) 1997-2000
// Utrecht University (The Netherlands),
// ETH Zurich (Switzerland),
// INRIA Sophia-Antipolis (France),
// Max-Planck-Institute Saarbruecken (Germany),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Kernel_d/include/CGAL/Kernel_d/Kernel_classesHd.h $
// $Id: include/CGAL/Kernel_d/Kernel_classesHd.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Michael Seel <seel@mpi-sb.mpg.de>

#ifndef CGAL_KERNEL_CLASSESHD_H
#define CGAL_KERNEL_CLASSESHD_H

namespace CGAL {

template <typename RT, typename LA> class PointHd2;
template <typename RT, typename LA> class VectorHd;
template <typename RT, typename LA> class DirectionHd;
template <typename RT, typename LA> class HyperplaneHd;
template <typename RT, typename LA> class Aff_transformationHd;

} //namespace CGAL

#endif //CGAL_KERNEL_CLASSESHD_H
