// Copyright (c) 2003,2004,2005,2006  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Segment_Delaunay_graph_2/include/CGAL/Segment_Delaunay_graph_2/Predicates_C2.h $
// $Id: include/CGAL/Segment_Delaunay_graph_2/Predicates_C2.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Menelaos Karavelas <mkaravel@iacm.forth.gr>




#ifndef CGAL_SEGMENT_DELAUNAY_GRAPH_2_PREDICATES_C2_H
#define CGAL_SEGMENT_DELAUNAY_GRAPH_2_PREDICATES_C2_H

#include <CGAL/license/Segment_Delaunay_graph_2.h>


#include <CGAL/Segment_Delaunay_graph_2/Voronoi_vertex_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Compare_x_2.h>
#include <CGAL/Segment_Delaunay_graph_2/Compare_y_2.h>
#include <CGAL/Segment_Delaunay_graph_2/Are_parallel_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Are_same_points_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Are_same_segments_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Orientation_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Oriented_side_of_bisector_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Vertex_conflict_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Finite_edge_interior_conflict_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Infinite_edge_interior_conflict_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Is_degenerate_edge_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Arrangement_type_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Arrangement_type_non_intersecting_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Oriented_side_C2.h>

#endif // CGAL_SEGMENT_DELAUNAY_GRAPH_2_PREDICATES_C2_H
