// Copyright (c) 2025  GeometryFactory Sarl (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Constrained_triangulation_3/include/CGAL/Conforming_constrained_Delaunay_triangulation_3_fwd.h $
// $Id: include/CGAL/Conforming_constrained_Delaunay_triangulation_3_fwd.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Laurent Rineau

#ifndef CGAL_CONFORMING_CONSTRAINED_DELAUNAY_TRIANGULATION_3_FWD_H
#define CGAL_CONFORMING_CONSTRAINED_DELAUNAY_TRIANGULATION_3_FWD_H

#include <CGAL/license/Constrained_triangulation_3.h>

namespace CGAL {

template <typename Traits, typename Tr>
class Conforming_constrained_Delaunay_triangulation_3;

} // namespace CGAL

#endif // CGAL_CONFORMING_CONSTRAINED_DELAUNAY_TRIANGULATION_3_FWD_H
