// Copyright (c) 2018-2019 GeometryFactory (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Optimal_bounding_box/include/CGAL/optimal_bounding_box.h $
// $Id: include/CGAL/optimal_bounding_box.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Konstantinos Katrioplas
//                 Mael Rouxel-Labbé
//
#ifndef CGAL_OPTIMAL_BOUNDING_BOX_OBB_H
#define CGAL_OPTIMAL_BOUNDING_BOX_OBB_H

#ifndef DOXYGEN_RUNNING
#include <CGAL/license/Optimal_bounding_box.h>
#endif

/**
* \ingroup PkgOptimalBoundingBoxRef
* \file CGAL/optimal_bounding_box.h
* Convenience header file including the headers for all the classes and functions of this package.
*/

#include <CGAL/Optimal_bounding_box/Oriented_bounding_box_traits_3.h>
#include <CGAL/Optimal_bounding_box/oriented_bounding_box.h>

#endif // CGAL_OPTIMAL_BOUNDING_BOX_OBB_H
