// Copyright (c) 2005-2008  INRIA Sophia-Antipolis (France).
//               2008 GeometryFactory
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/CGAL_ImageIO/include/CGAL/read_vtk_image_data.h $
// $Id: include/CGAL/read_vtk_image_data.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Laurent Rineau, Pierre Alliez

#ifndef CGAL_READ_VTK_IMAGE_DATA_H
#define CGAL_READ_VTK_IMAGE_DATA_H

#ifndef CGAL_NO_DEPRECATED_CODE
#include <CGAL/IO/read_vtk_image_data.h>

#define CGAL_DEPRECATED_HEADER "<CGAL/read_vtk_image_data.h>"
#define CGAL_REPLACEMENT_HEADER "<CGAL/IO/read_vtk_image_data.h>"
#include <CGAL/Installation/internal/deprecation_warning.h>
#endif

#endif // CGAL_READ_VTK_IMAGE_DATA_H
