// Copyright (c) 2000
// Utrecht University (The Netherlands),
// ETH Zurich (Switzerland),
// INRIA Sophia-Antipolis (France),
// Max-Planck-Institute Saarbruecken (Germany),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Intersections_2/include/CGAL/intersection_2.h $
// $Id: include/CGAL/intersection_2.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Geert-Jan Giezeman



#ifndef CGAL_INTERSECTION_2_H
#define CGAL_INTERSECTION_2_H

#include <CGAL/Intersections_2/Bbox_2_Circle_2.h>
#include <CGAL/Intersections_2/Bbox_2_Line_2.h>
#include <CGAL/Intersections_2/Bbox_2_Point_2.h>
#include <CGAL/Intersections_2/Bbox_2_Ray_2.h>

#include <CGAL/Intersections_2/Circle_2_Circle_2.h>
#include <CGAL/Intersections_2/Circle_2_Iso_rectangle_2.h>
#include <CGAL/Intersections_2/Circle_2_Line_2.h>
#include <CGAL/Intersections_2/Circle_2_Point_2.h>
#include <CGAL/Intersections_2/Circle_2_Segment_2.h>
#include <CGAL/Intersections_2/Circle_2_Ray_2.h>
#include <CGAL/Intersections_2/Circle_2_Triangle_2.h>


#include <CGAL/Intersections_2/Iso_rectangle_2_Iso_rectangle_2.h>
#include <CGAL/Intersections_2/Iso_rectangle_2_Line_2.h>
#include <CGAL/Intersections_2/Iso_rectangle_2_Point_2.h>
#include <CGAL/Intersections_2/Iso_rectangle_2_Ray_2.h>
#include <CGAL/Intersections_2/Iso_rectangle_2_Segment_2.h>
#include <CGAL/Intersections_2/Iso_rectangle_2_Triangle_2.h>

#include <CGAL/Intersections_2/Line_2_Line_2.h>
#include <CGAL/Intersections_2/Line_2_Point_2.h>
#include <CGAL/Intersections_2/Line_2_Ray_2.h>
#include <CGAL/Intersections_2/Line_2_Segment_2.h>
#include <CGAL/Intersections_2/Line_2_Triangle_2.h>

#include <CGAL/Intersections_2/Point_2_Point_2.h>
#include <CGAL/Intersections_2/Point_2_Ray_2.h>
#include <CGAL/Intersections_2/Point_2_Segment_2.h>
#include <CGAL/Intersections_2/Point_2_Triangle_2.h>

#include <CGAL/Intersections_2/Ray_2_Ray_2.h>
#include <CGAL/Intersections_2/Ray_2_Segment_2.h>
#include <CGAL/Intersections_2/Ray_2_Triangle_2.h>

#include <CGAL/Intersections_2/Segment_2_Segment_2.h>
#include <CGAL/Intersections_2/Segment_2_Triangle_2.h>

#include <CGAL/Intersections_2/Triangle_2_Triangle_2.h>

#endif // CGAL_INTERSECTION_2_H
