// Copyright (c) 2025 GeometryFactory Sarl (France).
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/STL_Extension/include/CGAL/unordered_flat_set.h $
// $Id: include/CGAL/unordered_flat_set.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Laurent Rineau

#ifndef CGAL_UNORDERED_FLAT_SET_H
#define CGAL_UNORDERED_FLAT_SET_H

// lazy implementation: define the two template aliases in the same header file
#include <CGAL/unordered_flat_map.h>

#endif // CGAL_UNORDERED_FLAT_SET_H
