// Copyright (c) 1997-2007  ETH Zurich (Switzerland).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/QP_solver/include/CGAL/QP_solver/basic.h $
// $Id: include/CGAL/QP_solver/basic.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Sven Schoenherr
//                 Bernd Gaertner <gaertner@inf.ethz.ch>
//                 Franz Wessendorp
//                 Kaspar Fischer

#ifndef CGAL_QP_SOLVER_BASIC_H
#define CGAL_QP_SOLVER_BASIC_H

#include <CGAL/license/QP_solver.h>


#include <CGAL/basic.h>
#include <CGAL/Quotient.h>
#include <CGAL/QP_solver/assertions.h>
#include <CGAL/QP_solver/debug.h>
#include <boost/functional.hpp>

#endif // CGAL_QP_SOLVER_BASIC_H

// ===== EOF ==================================================================
