// Copyright (c) 2017  GeometryFactory (France).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/BGL/include/CGAL/boost/graph/Graph_with_descriptor_with_graph_fwd.h $
// $Id: include/CGAL/boost/graph/Graph_with_descriptor_with_graph_fwd.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Laurent Rineau

#ifndef CGAL_BOOST_GRAPH_GWDWG_FWD_H
#define CGAL_BOOST_GRAPH_GWDWG_FWD_H

namespace CGAL
{
  template<typename Graph_>
  struct Graph_with_descriptor_with_graph;
}

#endif // CGAL_BOOST_GRAPH_GWDWG_FWD_H
